"""Exceptions for Forecast.Solar."""

from datetime import datetime
from typing import Any


class ForecastSolarError(Exception):
    """Generic Forecast.Solar exception."""


class ForecastSolarConnectionError(ForecastSolarError):
    """Forecast.Solar API connection exception."""


class ForecastSolarConfigError(ForecastSolarError):
    """Forecast.Solar API configuration exception."""

    def __init__(self, data: dict[str, str]) -> None:
        """Init a solar config error."""
        super().__init__(f"{data['text']} (error 422)")


class ForecastSolarAuthenticationError(ForecastSolarError):
    """Forecast.Solar API authentication exception."""

    def __init__(self, data: dict[str, str]) -> None:
        """Init a solar auth error.

        https://doc.forecast.solar/doku.php?id=api#invalid_request
        """
        # seems that code is missing in response in some endpoints (i.e /info)
        code = data.get("code")
        super().__init__(f"{data['text']} (error {code})")
        self.code = code


class ForecastSolarRequestError(ForecastSolarError):
    """Forecast.Solar wrong request input variables."""

    def __init__(self, data: dict[str, str]) -> None:
        """Init a solar request error.

        https://doc.forecast.solar/doku.php?id=api#invalid_request
        """
        super().__init__(f"{data['text']} (error {data['code']})")
        self.code = data["code"]


class ForecastSolarRatelimitError(ForecastSolarRequestError):
    """Forecast.Solar maximum number of requests reached exception."""

    def __init__(self, data: dict[str, Any]) -> None:
        """Init a rate limit error."""
        super().__init__(data)

        self.reset_at = datetime.fromisoformat(data["ratelimit"]["retry-at"])



"""Data models for the Forecast.Solar API."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import date, datetime, timedelta
from enum import Enum
from typing import TYPE_CHECKING, Any
from zoneinfo import ZoneInfo

if TYPE_CHECKING:
    from aiohttp import ClientResponse


def _timed_value(at: datetime, data: dict[datetime, int]) -> int | None:
    """Return the value for a specific time."""
    value = None
    for timestamp, cur_value in data.items():
        if timestamp > at:
            return value
        value = cur_value

    return None


def _interval_value_sum(
    interval_begin: datetime, interval_end: datetime, data: dict[datetime, int]
) -> int:
    """Return the sum of values in interval."""
    total = 0

    for timestamp, wh in data.items():
        # Skip all until this hour
        if timestamp < interval_begin:
            continue

        if timestamp >= interval_end:
            break

        total += wh

    return total


class AccountType(str, Enum):
    """Enumeration representing the Forecast.Solar account type."""

    PUBLIC = "public"
    PERSONAL = "personal"
    PROFESSIONAL = "professional"


@dataclass
class Estimate:
    """Object holding estimate forecast results from Forecast.Solar.

    Attributes
    ----------
        watts: Estimated solar power output per time period.
        wh_period: Estimated solar energy production differences per hour.
        wh_days: Estimated solar energy production per day.

    """

    watts: dict[datetime, int]
    wh_period: dict[datetime, int]
    wh_days: dict[datetime, int]
    api_rate_limit: int
    api_timezone: str

    @property
    def timezone(self) -> str:
        """Return API timezone information."""
        return self.api_timezone

    @property
    def account_type(self) -> AccountType:
        """Return API account_type information."""
        if self.api_rate_limit == 60:
            return AccountType.PERSONAL
        if self.api_rate_limit == 5:
            return AccountType.PROFESSIONAL
        return AccountType.PUBLIC

    @property
    def energy_production_today(self) -> int:
        """Return estimated energy produced today."""
        return self.day_production(self.now().date())

    @property
    def energy_production_tomorrow(self) -> int:
        """Return estimated energy produced today."""
        return self.day_production(self.now().date() + timedelta(days=1))

    @property
    def energy_production_today_remaining(self) -> int:
        """Return estimated energy produced in rest of today."""
        return _interval_value_sum(
            self.now(),
            self.now().replace(hour=0, minute=0, second=0, microsecond=0)
            + timedelta(days=1),
            self.wh_period,
        )

    @property
    def power_production_now(self) -> int:
        """Return estimated power production right now."""
        return self.power_production_at_time(self.now())

    @property
    def power_highest_peak_time_today(self) -> datetime:
        """Return datetime with highest power production moment today."""
        return self.peak_production_time(self.now().date())

    @property
    def power_highest_peak_time_tomorrow(self) -> datetime:
        """Return datetime with highest power production moment tomorrow."""
        return self.peak_production_time(self.now().date() + timedelta(days=1))

    @property
    def energy_current_hour(self) -> int:
        """Return the estimated energy production for the current hour."""
        return _interval_value_sum(
            self.now().replace(minute=0, second=0, microsecond=0),
            self.now().replace(minute=0, second=0, microsecond=0) + timedelta(hours=1),
            self.wh_period,
        )

    def day_production(self, specific_date: date) -> int:
        """Return the day production."""
        for timestamp, production in self.wh_days.items():
            if timestamp.date() == specific_date:
                return production

        return 0

    def now(self) -> datetime:
        """Return the current timestamp in the API timezone."""
        return datetime.now(tz=ZoneInfo(self.api_timezone))

    def peak_production_time(self, specific_date: date) -> datetime:
        """Return the peak time on a specific date."""
        value = max(
            (watt for date, watt in self.watts.items() if date.date() == specific_date),
            default=None,
        )
        for timestamp, watt in self.watts.items():
            if watt == value:
                return timestamp
        raise RuntimeError("No peak production time found")

    def power_production_at_time(self, time: datetime) -> int:
        """Return estimated power production at a specific time."""
        return _timed_value(time, self.watts) or 0

    def sum_energy_production(self, period_hours: int) -> int:
        """Return the sum of the energy production."""
        now = self.now().replace(minute=59, second=59, microsecond=999)
        until = now + timedelta(hours=period_hours)

        return _interval_value_sum(now, until, self.wh_period)

    @classmethod
    def from_dict(cls: type[Estimate], data: dict[str, Any]) -> Estimate:
        """Return a Estimate object from a Forecast.Solar API response.

        Converts a dictionary, obtained from the Forecast.Solar API into
        a Estimate object.

        Args:
        ----
            data: The estimate response from the Forecast.Solar API.

        Returns:
        -------
            An Estimate object.

        """
        return cls(
            watts={
                datetime.fromisoformat(d): w for d, w in data["result"]["watts"].items()
            },
            wh_period={
                datetime.fromisoformat(d): e
                for d, e in data["result"]["watt_hours_period"].items()
            },
            wh_days={
                datetime.fromisoformat(d): e
                for d, e in data["result"]["watt_hours_day"].items()
            },
            api_rate_limit=data["message"]["ratelimit"]["limit"],
            api_timezone=data["message"]["info"]["timezone"],
        )


@dataclass
class Ratelimit:
    """Information about the current rate limit."""

    call_limit: int
    remaining_calls: int
    period: int
    retry_at: datetime | None

    @classmethod
    def from_response(cls: type[Ratelimit], response: ClientResponse) -> Ratelimit:
        """Initialize rate limit object from response."""
        # The documented headers do not match the returned headers
        # https://doc.forecast.solar/doku.php?id=api#headers
        limit = int(response.headers["X-Ratelimit-Limit"])
        period = int(response.headers["X-Ratelimit-Period"])

        # Remaining is not there if we exceeded limit
        remaining = int(response.headers.get("X-Ratelimit-Remaining", 0))

        if "X-Ratelimit-Retry-At" in response.headers:
            retry_at = datetime.fromisoformat(response.headers["X-Ratelimit-Retry-At"])
        else:
            retry_at = None

        return cls(limit, remaining, period, retry_at)



"""Asynchronous Python client for the Forecast.Solar API."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Self

from aiodns import DNSResolver
from aiodns.error import DNSError
from aiohttp import ClientSession
from yarl import URL

from .exceptions import (
    ForecastSolarAuthenticationError,
    ForecastSolarConfigError,
    ForecastSolarConnectionError,
    ForecastSolarError,
    ForecastSolarRatelimitError,
    ForecastSolarRequestError,
)
from .models import Estimate, Ratelimit


@dataclass
class ForecastSolar:
    """Main class for handling connections with the Forecast.Solar API."""

    azimuth: float
    declination: float
    kwp: float
    latitude: float
    longitude: float

    api_key: str | None = None
    damping: float = 0
    damping_morning: float | None = None
    damping_evening: float | None = None
    horizon: str | None = None

    session: ClientSession | None = None
    ratelimit: Ratelimit | None = None
    inverter: float | None = None
    _close_session: bool = False

    async def _request(
        self,
        uri: str,
        *,
        rate_limit: bool = True,
        authenticate: bool = True,
        params: dict[str, Any] | None = None,
    ) -> Any:
        """Handle a request to the Forecast.Solar API.

        A generic method for sending/handling HTTP requests done against
        the Forecast.Solar API.

        Args:
        ----
            uri: Request URI, for example, 'estimate'
            rate_limit: Parse rate limit from response. Set to False for
                endpoints that are missing rate limiting headers in response.
            authenticate: Prefix request with api_key. Set to False for
                endpoints that do not provide authentication.

        Returns:
        -------
            A Python dictionary (JSON decoded) with the response from
            the Forecast.Solar API.

        Raises:
        ------
            ForecastSolarAuthenticationError: If the API key is invalid.
            ForecastSolarConnectionError: An error occurred while communicating
                with the Forecast.Solar API.
            ForecastSolarError: Received an unexpected response from the
                Forecast.Solar API.
            ForecastSolarRequestError: There is something wrong with the
                variables used in the request.
            ForecastSolarRatelimitError: The number of requests has exceeded
                the rate limit of the Forecast.Solar API.

        """
        # Forecast.Solar is currently experiencing IPv6 issues.
        # However, their DNS does return an non-working IPv6 address.
        # This ensures we use the IPv4 address.
        dns = DNSResolver()
        try:
            result = await dns.query("api.forecast.solar", "A")
        except DNSError as err:
            raise ForecastSolarConnectionError(
                "Error while resolving Forecast.Solar API address"
            ) from err

        if not result:
            raise ForecastSolarConnectionError(
                "Could not resolve Forecast.Solar API address"
            )

        # Connect as normal
        url = URL.build(scheme="https", host=result[0].host)

        # Add API key if one is provided
        if authenticate and self.api_key is not None:
            url = url.with_path(f"{self.api_key}/")

        url = url.join(URL(uri))

        if self.session is None:
            self.session = ClientSession()
            self._close_session = True

        response = await self.session.request(
            "GET",
            url,
            params=params,
            headers={"Host": "api.forecast.solar"},
            ssl=False,
        )

        if response.status in (502, 503):
            raise ForecastSolarConnectionError("The Forecast.Solar API is unreachable")

        if response.status == 400:
            data = await response.json()
            raise ForecastSolarRequestError(data["message"])

        if response.status in (401, 403):
            data = await response.json()
            raise ForecastSolarAuthenticationError(data["message"])

        if response.status == 422:
            data = await response.json()
            raise ForecastSolarConfigError(data["message"])

        if response.status == 429:
            data = await response.json()
            raise ForecastSolarRatelimitError(data["message"])

        if rate_limit and response.status == 200:
            self.ratelimit = Ratelimit.from_response(response)

        response.raise_for_status()

        content_type = response.headers.get("Content-Type", "")
        if "application/json" not in content_type:
            text = await response.text()
            raise ForecastSolarError(
                "Unexpected response from the Forecast.Solar API",
                {"Content-Type": content_type, "response": text},
            )

        return await response.json()

    async def validate_plane(self) -> bool:
        """Validate plane by calling the Forecast.Solar API.

        Returns
        -------
            True, if plane is valid.

        """
        await self._request(
            f"check/{self.latitude}/{self.longitude}"
            f"/{self.declination}/{self.azimuth}/{self.kwp}",
            rate_limit=False,
            authenticate=False,
        )

        return True

    async def validate_api_key(self) -> bool:
        """Validate api key by calling the Forecast.Solar API.

        Returns
        -------
            True, if api key is valid

        """
        await self._request("info", rate_limit=False)

        return True

    async def estimate(self, actual: float = 0) -> Estimate:
        """Get solar production estimations from the Forecast.Solar API.

        Args:
        ----
            actual: The production for the day in kWh so far. Used to improve
                the estimation for the current day if an API key is provided.

        Returns:
        -------
            A Estimate object, with a estimated production forecast.

        """
        params = {"time": "utc", "damping": str(self.damping)}
        if self.inverter is not None:
            params["inverter"] = str(self.inverter)
        if self.horizon is not None:
            params["horizon"] = str(self.horizon)
        if self.damping_morning is not None and self.damping_evening is not None:
            params["damping_morning"] = str(self.damping_morning)
            params["damping_evening"] = str(self.damping_evening)
        if self.api_key is not None:
            params["actual"] = str(actual)

        data = await self._request(
            f"estimate/{self.latitude}/{self.longitude}"
            f"/{self.declination}/{self.azimuth}/{self.kwp}",
            params=params,
        )

        return Estimate.from_dict(data)

    async def close(self) -> None:
        """Close open client session."""
        if self.session and self._close_session:
            await self.session.close()

    async def __aenter__(self) -> Self:
        """Async enter.

        Returns
        -------
            The ForecastSolar object.

        """
        return self

    async def __aexit__(self, *_exc_info: object) -> None:
        """Async exit.

        Args:
        ----
            _exc_info: Exec type.

        """
        await self.close()



"""Asynchronous Python client for the Forecast.Solar API."""

from .exceptions import (
    ForecastSolarAuthenticationError,
    ForecastSolarConfigError,
    ForecastSolarConnectionError,
    ForecastSolarError,
    ForecastSolarRatelimitError,
    ForecastSolarRequestError,
)
from .forecast_solar import ForecastSolar
from .models import AccountType, Estimate, Ratelimit

__all__ = [
    "AccountType",
    "Estimate",
    "ForecastSolar",
    "ForecastSolarAuthenticationError",
    "ForecastSolarConfigError",
    "ForecastSolarConnectionError",
    "ForecastSolarError",
    "ForecastSolarRatelimitError",
    "ForecastSolarRequestError",
    "Ratelimit",
]
